from time import monotonic, sleep

from .client import RPCClient
from .exceptions import HungNodeError
from .utils import blockid_to_blocknumber, get_first_or_none


class SteemdClient(RPCClient):

    def get_block(self, block_number=None):
        if block_number is None:
            block_number = self.head_block_number

        return self.call('condenser_api.get_block', block_number)

    def get_blocks(self, start_block_number, end_block_number=None):
        if end_block_number is None:
            end_block_number = self.head_block_number

        for block_number in range(start_block_number, end_block_number + 1):
            yield self.get_block(block_number)

    def stream_blocks(self, irreversible=True, interval=None, max_blocks_catchup=None, node_hung_treshold=None):
        previous_block_nr = self.last_irreversible_block_number if irreversible else self.head_block_number
        if interval is None:
            interval = self.block_interval
        same_block_count = 0
        while True:
            t_start = monotonic()
            end_block_nr = self.last_irreversible_block_number if irreversible else self.head_block_number

            if node_hung_treshold is not None:
                if end_block_nr == previous_block_nr:
                    same_block_count += 1
                    if same_block_count > node_hung_treshold:
                        raise HungNodeError
                elif end_block_nr > previous_block_nr:
                    same_block_count = 0

            if max_blocks_catchup is not None:
                previous_block_nr = max(previous_block_nr, (end_block_nr - max_blocks_catchup))

            for block in self.get_blocks(start_block_number=previous_block_nr + 1, end_block_number=end_block_nr):
                yield block
                if block is not None:
                    previous_block_nr = blockid_to_blocknumber(block['block_id'])

            sleep_time = max((interval - (monotonic() - t_start)), 0.1)
            sleep(sleep_time)

    def get_account(self, account):
        return get_first_or_none(self.get_accounts(account))

    def get_accounts(self, accounts):
        if isinstance(accounts, str):
            accounts = [accounts]

        return self.call('condenser_api.get_accounts', accounts)

    def get_account_reputation(self, account):
        reputation = get_first_or_none(self.call('condenser_api.get_account_reputations', account, 1))
        try:
            reputation = int(reputation['reputation'])
        except (TypeError, KeyError, ValueError):
            reputation = None

        return reputation

    def get_account_reputations(self, accounts):
        return [self.get_account_reputation(account) for account in accounts]

    def get_witnesses_by_id(self, ids):
        if isinstance(ids, str):
            ids = [ids]

        return self.call('condenser_api.get_witnesses', ids)

    def get_witnesses_by_account(self, accounts):
        if isinstance(accounts, str):
            accounts = [accounts]

        witnesses = []
        for account in accounts:
            witnesses.append(self.call('condenser_api.get_witness_by_account', account))

        return witnesses

    @property
    def dynamic_global_properties(self):
        return self.call('condenser_api.get_dynamic_global_properties')

    @property
    def chain_properties(self):
        return self.call('condenser_api.get_chain_properties')

    @property
    def config(self):
        return self.call('condenser_api.get_config')

    @property
    def last_irreversible_block_number(self):
        return self.dynamic_global_properties.get('last_irreversible_block_num')

    @property
    def head_block_number(self):
        return self.dynamic_global_properties.get('head_block_number')

    @property
    def block_interval(self):
        # We cache this
        if not hasattr(self, '_block_interval'):
            self._block_interval = self.config.get('STEEM_BLOCK_INTERVAL', 3)

        return self._block_interval



from itertools import cycle
import json
import logging
from socket import SOL_SOCKET, SO_KEEPALIVE

import certifi
from urllib3.connection import HTTPSConnection
from urllib3.connectionpool import connection_from_url
from urllib3.exceptions import TimeoutError, MaxRetryError, ClosedPoolError
from urllib3.response import HTTPResponse
from urllib3.util.retry import Retry
from urllib3.util.timeout import Timeout


logger = logging.getLogger(__name__)


METHOD_WHITELIST = Retry.DEFAULT_METHOD_WHITELIST.union({'POST'})
STATUS_FORCELIST = frozenset({408, 444, 499, 500, 502, 503, 504, 520, 521, 522, 523, 524, 527})
DEFAULT_RETRY_OPTS = {
    'connect': 5,
    'read': 0,
    'redirect': 5,
    'status': 3,
    'method_whitelist': METHOD_WHITELIST,
    'status_forcelist': STATUS_FORCELIST,
    'backoff_factor': 0.1,
    'raise_on_redirect': True,
    'raise_on_status': True
}


class RPCClient(object):

    def __init__(self, nodes, **kwargs):
        if 'retries' in kwargs:
            # We need to make sure these are set the way we want.
            retry = kwargs['retries']
            retry_params = {
                'method_whitelist': set(retry.method_whitelist).add('POST'),
                'raise_on_redirect': True,
                'raise_on_status': True,
            }
            kwargs['retries'] = retry.new(**retry_params)

        self.connection_options = {
            'maxsize': 5,
            'timeout': Timeout(total=5.0),
            'retries': Retry(**DEFAULT_RETRY_OPTS),
            'cert_reqs': 'CERT_REQUIRED',
            'ca_certs': certifi.where(),
            'headers': {
                'Cache-Control': 'no-cache',
                'Connection': 'keep-alive',
                'Content-Type': 'application/json',
            },
            'socket_options': HTTPSConnection.default_socket_options + [(SOL_SOCKET, SO_KEEPALIVE, 1)]
        }
        self.connection_options.update(kwargs)
        self._connection = None

        self.failover_limit = len(nodes)

        self._nodes = cycle(nodes)
        self.connect()

    def __del__(self):
        self.disconnect()

    @property
    def nodes(self):
        nodes = []
        while True:
            node = next(self._nodes)
            if node in nodes:
                break
            nodes.append(node)

        return frozenset(nodes)

    @property
    def current_node(self):
        return self._connection.host

    def next_node(self):
        return self.connect(next(self._nodes))

    def connect(self, node=None):
        node_url = node or next(self._nodes)
        self.disconnect()
        self._connection = connection_from_url(node_url, **self.connection_options)
        logger.info("Current node changed to `{}`".format(self._connection.host))

        return node_url

    def disconnect(self):
        if self._connection:
            try:
                self._connection.close()
            except AttributeError:
                # This is a temporary workaround for https://github.com/shazow/urllib3/pull/1279
                pass

    def exec(self, name, *args, **kwargs):
        """Merely here for backwards-compatibility
        """
        return self.call(name, *args, **kwargs)

    def call(self, proc, *args, **kwargs):
        failover_limit = self.failover_limit if kwargs.pop('enable_failover', True) else 0
        failovers = 0
        response = None
        while failovers <= failover_limit:
            try:
                response = self._call(proc, *args, **kwargs)
            except ClosedPoolError:
                logger.exception("Trying to make a call, but the connection pool is closed. Trying to reconnect.")
                self.connect()
            except MaxRetryError as e:
                logger.exception(e.reason)
                if failovers < failover_limit:
                    self.next_node()
            else:
                if isinstance(response, HTTPResponse):
                    break

            failovers += 1

        decoded_response = {}
        try:
            decoded_response = json.loads(response.data.decode('utf-8'))
        except AttributeError:
            logger.exception("Unable to get a response when calling remote procedure `{}`".format(proc))
        except (UnicodeError, json.JSONDecodeError):
            logger.exception("Unable to decode response.")
        else:
            if not isinstance(decoded_response, dict):
                logger.critical("Invalid response received.")
                decoded_response = {}

        if 'error' in decoded_response:
            logger.critical(decoded_response['error'].get('message', 'Unknown RPC error occurred.'))

        return decoded_response.get('result', None)

    def _call(self, proc, *args, **kwargs):
        """This method should either return a valid response, or raise an exception.
        """
        body = {
            'jsonrpc': '2.0',
            'id': kwargs.pop('call_id', 0),
        }
        api = kwargs.pop('api', None)
        body['params'] = kwargs if kwargs else args
        if api:
            body['method'] = "{api}.{function}".format(api=api, function=proc)
        else:
            body['method'] = proc

        body = json.dumps(body, ensure_ascii=False, default=str, separators=(',', ':')).encode('utf-8')
        return self._connection.request('POST', '/', body=body)



#!/usr/bin/env python3

from setuptools import setup, find_packages
import sys

from nozzle import __version__ as version


if sys.version_info < (3, 5):
    sys.exit('Only Python >=3.5 supported.')


setup(
    name='steem-nozzle',
    version=version,
    description='A Python library for the Steem Blockchain',
    long_description='A minimalistic Python client library for the Steem Blockchain',
    url='https://github.com/blockbrothers/nozzle',
    author='blockbrothers',
    author_email='info@blockbrothers.io',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='steem blockchain',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='>=3.5',
    install_requires=['urllib3', 'certifi'],
    extras_require={},
    package_data={},
    entry_points={},
)
